from typing import TYPE_CHECKING

from django.db.models import QuerySet

from reversion import create_revision, set_comment

from aleksis.core.managers import AlekSISBaseManagerWithoutMigrations, DateRangeQuerySetMixin

if TYPE_CHECKING:
    from .models import BreakSlot, Lesson, Slot, Supervision, TimeGrid, ValidityRange


class ValidityRangeQuerySet(QuerySet, DateRangeQuerySetMixin):
    """Custom query set for validity ranges."""


class ValidityRangeManager(AlekSISBaseManagerWithoutMigrations):
    """Manager for validity ranges."""

    def copy_from_to(
        self, source: ValidityRange, target: ValidityRange
    ) -> dict[str, dict[any, any]]:
        from .models import Lesson, Slot, Supervision, TimeGrid

        time_grid_obj_map = TimeGrid.objects.copy_from_to(source, target)

        slot_map = {}
        for time_grid_source, time_grid_target in time_grid_obj_map.items():
            slot_map |= Slot.objects.copy_from_to(time_grid_source, time_grid_target)

        lesson_obj_map = Lesson.objects.copy_from_to(source, slot_map)

        supervision_obj_map = Supervision.objects.copy_from_to(source, slot_map)

        obj_map = dict(
            time_grid=time_grid_obj_map,
            slot=slot_map,
            lesson=lesson_obj_map,
            supervision=supervision_obj_map,
        )
        return obj_map


class TimeGridManager(AlekSISBaseManagerWithoutMigrations):
    """Manager for time grids."""

    def copy_from_to(
        self, source: ValidityRange, target: ValidityRange
    ) -> list[tuple[TimeGrid, TimeGrid]]:
        obj_map = []
        with create_revision():
            set_comment(f"Copy time grids from validity range {source.pk} to {target.pk}")

            for time_grid in source.time_grids.all():
                new_time_grid = self.create(validity_range=self, group=time_grid.group)

                obj_map.append((time_grid, new_time_grid))

        return obj_map


class SlotManager(AlekSISBaseManagerWithoutMigrations):
    """Manager for slots."""

    def copy_from_to(
        self, source: TimeGrid, target: TimeGrid
    ) -> [dict[Slot, Slot], QuerySet[Slot]]:
        obj_map = {}
        with create_revision():
            set_comment(f"Copy slots from time grid {source.pk} to {target.pk}")
            slots = self.filter(time_grid=source)

            update_or_created_objs = []

            for slot in slots:
                slot: BreakSlot | Slot
                klass = slot.get_real_instance_class()

                defaults = {
                    "name": slot.name,
                    "time_start": slot.time_start,
                    "time_end": slot.time_end,
                }

                obj, __ = klass.objects.update_or_create(
                    weekday=slot.weekday, time_grid=target, period=slot.period, defaults=defaults
                )
                obj_map[slot] = obj
                update_or_created_objs.append(obj)

            # Delete all slots in the time_grid that are not in the from_time_grid
            objects_to_delete = self.filter(time_grid=target).exclude(id__in=update_or_created_objs)
            objects_to_delete.delete()

        return obj_map, objects_to_delete


class LessonManager(AlekSISBaseManagerWithoutMigrations):
    def copy_from_to(
        self, source: ValidityRange, target_slot_map: dict[Slot, Slot]
    ) -> dict[Lesson, Lesson]:
        obj_map = {}

        lessons = self.filter(slot_start__validity_range=source)
        for lesson in lessons:
            new_slot_start = target_slot_map.get(lesson.slot_start)
            new_slot_end = target_slot_map.get(lesson.slot_end)
            if not new_slot_start or not new_slot_end:
                continue

            new_lesson = self.create(
                course=lesson.course,
                slot_start=new_slot_start,
                slot_end=new_slot_end,
                recurrence=lesson.recurrence,
                subject=lesson.subject,
            )
            new_lesson.rooms.set(lesson.rooms.all())
            new_lesson.teachers.set(lesson.teachers.all())

            obj_map[lesson] = new_lesson

        return obj_map


class SupervisionManager(AlekSISBaseManagerWithoutMigrations):
    def copy_from_to(
        self, source: ValidityRange, target_slot_map: dict[Slot, Slot]
    ) -> dict[Supervision, Supervision]:
        obj_map = {}

        supervisions = self.filter(break_slot__validity_range=source)
        for supervision in supervisions:
            new_slot = target_slot_map.get(supervision.break_slot)
            if not new_slot:
                continue

            new_supervision = self.create(
                break_slot=supervision.break_slot,
                recurrence=supervision.recurrence,
                subject=supervision.subject,
            )
            new_supervision.rooms.set(supervision.rooms.all())
            new_supervision.teachers.set(supervision.teachers.all())

            obj_map[supervision] = new_supervision

        return obj_map
