from django.core.exceptions import PermissionDenied

import graphene
from graphene_django.types import DjangoObjectType
from graphene_django_cud.mutations import (
    DjangoBatchDeleteMutation,
    DjangoBatchPatchMutation,
    DjangoCreateMutation,
)
from guardian.shortcuts import get_objects_for_user
from reversion import create_revision, set_comment, set_user

from aleksis.core.schema.base import (
    DeleteMutation,
    DjangoFilterMixin,
    PermissionBatchDeleteMixin,
    PermissionBatchPatchMixin,
    PermissionsTypeMixin,
)

from ..models import TimeGrid, ValidityRange


class ValidityRangeType(PermissionsTypeMixin, DjangoFilterMixin, DjangoObjectType):
    class Meta:
        model = ValidityRange
        fields = ("id", "school_term", "name", "date_start", "date_end", "status", "time_grids")
        filter_fields = {
            "id": ["exact"],
            "school_term": ["exact", "in"],
            "status": ["exact"],
            "name": ["icontains", "exact"],
            "date_start": ["exact", "lt", "lte", "gt", "gte"],
            "date_end": ["exact", "lt", "lte", "gt", "gte"],
        }

    @classmethod
    def get_queryset(cls, queryset, info):
        if not info.context.user.has_perm("lesrooster.view_validityrange_rule"):
            return get_objects_for_user(
                info.context.user, "lesrooster.view_validityrange", queryset
            )
        return queryset


class ValidityRangeCreateMutation(DjangoCreateMutation):
    class Meta:
        model = ValidityRange
        permissions = ("lesrooster.create_validity_range_rule",)
        only_fields = (
            "id",
            "school_term",
            "name",
            "date_start",
            "date_end",
            "status",
            "time_grids",
        )
        field_types = {"status": graphene.String()}


class ValidityRangeDeleteMutation(DeleteMutation):
    klass = ValidityRange
    permission_required = "lesrooster.delete_validityrange_rule"


class ValidityRangeBatchDeleteMutation(PermissionBatchDeleteMixin, DjangoBatchDeleteMutation):
    class Meta:
        model = ValidityRange
        permissions = ("lesrooster.delete_validityrange_rule",)


class ValidityRangeBatchPatchMutation(PermissionBatchPatchMixin, DjangoBatchPatchMutation):
    class Meta:
        model = ValidityRange
        permissions = ("lesrooster.change_validityrange",)
        only_fields = (
            "id",
            "school_term",
            "name",
            "date_start",
            "date_end",
            "status",
            "time_grids",
        )
        field_types = {"status": graphene.String()}


class CopyDataFromValidityRangeMutation(graphene.Mutation):
    class Arguments:
        from_validty_range = graphene.ID()
        to_validity_range = graphene.ID()

    result = graphene.Boolean()

    @classmethod
    def mutate(cls, root, info, from_validity_range, to_validity_range):
        if not info.context.user.has_perm("lesrooster.edit_slot_rule"):
            raise PermissionDenied()

        from_validity_range = TimeGrid.objects.get(id=from_validity_range)
        to_validity_range = TimeGrid.objects.get(id=to_validity_range)

        with create_revision():
            set_user(info.context.user)
            set_comment(
                f"Copy data from validity range {from_validity_range.pk} to {to_validity_range.pk}"
            )
            to_validity_range.clear()
            ValidityRange.objects.copy_from_to(from_validity_range, to_validity_range)

        return cls(result=True)
